@@ -31,7 +31,10 @@
 #ifndef CAIROFONTENGINE_H
 #define CAIROFONTENGINE_H
 
+#include <memory>
 #include <mutex>
+#include <unordered_map>
+#include <vector>
 
 #include "poppler-config.h"
 #include <cairo-ft.h>
@@ -44,7 +47,7 @@ class CairoFontEngine;
 class CairoFont
 {
 public:
-    CairoFont(Ref refA, cairo_font_face_t *cairo_font_faceA, int *codeToGIDA, unsigned int codeToGIDLenA, bool substituteA, bool printingA);
+    CairoFont(Ref refA, cairo_font_face_t *cairo_font_faceA, std::vector<int> &&codeToGIDA, bool substituteA, bool printingA);
     virtual ~CairoFont();
     CairoFont(const CairoFont &) = delete;
     CairoFont &operator=(const CairoFont &other) = delete;
@@ -52,7 +55,7 @@ public:
     virtual bool matches(Ref &other, bool printing);
     cairo_font_face_t *getFontFace();
     unsigned long getGlyph(CharCode code, const Unicode *u, int uLen);
-    double getSubstitutionCorrection(GfxFont *gfxFont);
+    double getSubstitutionCorrection(const std::shared_ptr<GfxFont> &gfxFont);
 
     bool isSubstitute() { return substitute; }
 
@@ -60,8 +63,7 @@ protected:
     Ref ref;
     cairo_font_face_t *cairo_font_face;
 
-    int *codeToGID;
-    unsigned int codeToGIDLen;
+    std::vector<int> codeToGID;
 
     bool substitute;
     bool printing;
@@ -69,14 +71,22 @@ protected:
 
 //------------------------------------------------------------------------
 
+struct FreeTypeFontFace
+{
+    FT_Face face;
+    cairo_font_face_t *cairo_font_face;
+};
+
 class CairoFreeTypeFont : public CairoFont
 {
 public:
-    static CairoFreeTypeFont *create(GfxFont *gfxFont, XRef *xref, FT_Library lib, bool useCIDs);
+    static CairoFreeTypeFont *create(const std::shared_ptr<GfxFont> &gfxFont, XRef *xref, FT_Library lib, CairoFontEngine *fontEngine, bool useCIDs);
     ~CairoFreeTypeFont() override;
 
 private:
-    CairoFreeTypeFont(Ref ref, cairo_font_face_t *cairo_font_face, int *codeToGID, unsigned int codeToGIDLen, bool substitute);
+    CairoFreeTypeFont(Ref ref, cairo_font_face_t *cairo_font_face, std::vector<int> &&codeToGID, bool substitute);
+
+    static std::optional<FreeTypeFontFace> getFreeTypeFontFace(CairoFontEngine *fontEngine, FT_Library lib, const std::string &filename, std::vector<unsigned char> &&data);
 };
 
 //------------------------------------------------------------------------
@@ -84,19 +94,17 @@ private:
 class CairoType3Font : public CairoFont
 {
 public:
-    static CairoType3Font *create(const std::shared_ptr<const GfxFont> &gfxFont, PDFDoc *doc, CairoFontEngine *fontEngine, bool printing, XRef *xref);
+    static CairoType3Font *create(const std::shared_ptr<GfxFont> &gfxFont, PDFDoc *doc, CairoFontEngine *fontEngine, bool printing, XRef *xref);
     ~CairoType3Font() override;
 
     bool matches(Ref &other, bool printing) override;
 
 private:
-    CairoType3Font(Ref ref, cairo_font_face_t *cairo_font_face, int *codeToGID, unsigned int codeToGIDLen, bool printing, XRef *xref);
+    CairoType3Font(Ref ref, cairo_font_face_t *cairo_font_face, std::vector<int> &&codeToGIDA, bool printing, XRef *xref);
 };
 
 //------------------------------------------------------------------------
 
-#define cairoFontCacheSize 64
-
 //------------------------------------------------------------------------
 // CairoFontEngine
 //------------------------------------------------------------------------
@@ -110,13 +118,23 @@ public:
     CairoFontEngine(const CairoFontEngine &) = delete;
     CairoFontEngine &operator=(const CairoFontEngine &other) = delete;
 
-    CairoFont *getFont(const std::shared_ptr<GfxFont> &gfxFont, PDFDoc *doc, bool printing, XRef *xref);
+    std::shared_ptr<CairoFont> getFont(const std::shared_ptr<GfxFont> &gfxFont, PDFDoc *doc, bool printing, XRef *xref);
+
+    static std::optional<FreeTypeFontFace> getExternalFontFace(FT_Library ftlib, const std::string &filename);
 
 private:
-    CairoFont *fontCache[cairoFontCacheSize];
     FT_Library lib;
     bool useCIDs;
-    mutable std::recursive_mutex mutex;
+    mutable std::mutex mutex;
+
+    // Cache of CairoFont for current document
+    // Most recently used is at the end of the vector.
+    static const size_t cairoFontCacheSize = 64;
+    std::vector<std::shared_ptr<CairoFont>> fontCache;
+
+    // Global cache of cairo_font_face_t for external font files.
+    static std::unordered_map<std::string, FreeTypeFontFace> fontFileCache;
+    static std::recursive_mutex fontFileCacheMutex;
 };
 
 #endif
